const Chunk = @import("./Chunk.zig");

const Expr = @import("./Expr.zig");

const Stmt = @import("./Stmt.zig");

const coral = @import("coral");

const kym = @import("../kym.zig");

const tree = @import("./tree.zig");

chunk: *Chunk,
env: *kym.RuntimeEnv,

const Self = @This();

fn compile_argument(self: Self, environment: *const tree.Environment, initial_argument: ?*const Expr) kym.RuntimeError!u8 {
	// TODO: Exceeding 255 arguments will make the VM crash.
	var maybe_argument = initial_argument;
	var argument_count = @as(u8, 0);

	while (maybe_argument) |argument| {
		try self.compile_expression(environment, argument);

		maybe_argument = argument.next;
		argument_count += 1;
	}

	return argument_count;
}

fn compile_expression(self: Self, environment: *const tree.Environment, expression: *const Expr) kym.RuntimeError!void {
	const number_format = coral.utf8.DecimalFormat{
		.delimiter = "_",
		.positive_prefix = .none,
	};

	switch (expression.kind) {
		.nil_literal => try self.chunk.opcodes.push_one(.push_nil),
		.true_literal => try self.chunk.opcodes.push_one(.push_true),
		.false_literal => try self.chunk.opcodes.push_one(.push_false),

		.number_literal => |literal| {
			for (literal) |codepoint| {
				if (codepoint == '.') {
					return self.chunk.opcodes.push_one(.{
						.push_const = try self.declare_float(number_format.parse(literal, kym.Float) orelse unreachable),
					});
				}
			}

			try self.chunk.opcodes.push_one(.{
				.push_const = try self.declare_fixed(number_format.parse(literal, kym.Fixed) orelse unreachable),
			});
		},

		.string_literal => |literal| {
			try self.chunk.opcodes.push_one(.{.push_const = try self.declare_string(literal)});
		},

		.symbol_literal => |literal| {
			try self.chunk.opcodes.push_one(.{.push_const = try self.declare_symbol(literal)});
		},

		.table_construct => |table_construct| {
			var table_entry = table_construct.entry;
			var field_count = @as(u32, 0);

			while (table_entry) |entry| : (table_entry = entry.next) {
				try self.compile_expression(environment, entry);

				if (entry.kind != .key_value) {
					try self.chunk.opcodes.push_one(.push_top);
				}

				field_count += 1;
			}

			try self.chunk.opcodes.push_one(.{.push_table = field_count});
		},

		.key_value => |key_value| {
			try self.compile_expression(environment, key_value.value);
			try self.compile_expression(environment, key_value.key);
		},

		.lambda_construct => |lambda_construct| {
			var chunk = try Chunk.make(self.env, "<lambda>", lambda_construct.environment);

			errdefer chunk.free(self.env);

			if (lambda_construct.environment.capture_count == 0 and environment.capture_count == 0) {
				try self.chunk.opcodes.push_one(.{.push_const = try self.declare_chunk(chunk)});
			} else {
				for (environment.captures[0 .. environment.capture_count]) |capture| {
					try self.chunk.opcodes.push_one(.{.push_local = environment.captures[capture]});
				}

				try self.chunk.opcodes.push_one(.{.push_const = try self.declare_chunk(chunk)});

				try self.chunk.opcodes.push_one(.{
					.bind = lambda_construct.environment.capture_count + environment.capture_count,
				});
			}
		},

		.binary_op => |binary_op| {
			try self.compile_expression(environment, binary_op.lhs_operand);
			try self.compile_expression(environment, binary_op.rhs_operand);

			try self.chunk.opcodes.push_one(switch (binary_op.operation) {
				.addition => .add,
				.subtraction => .sub,
				.multiplication => .mul,
				.divsion => .div,
				.greater_equals_comparison => .cge,
				.greater_than_comparison => .cgt,
				.equals_comparison => .eql,
				.less_than_comparison => .clt,
				.less_equals_comparison => .cle,
			});
		},

		.unary_op => |unary_op| {
			try self.compile_expression(environment, unary_op.operand);

			try self.chunk.opcodes.push_one(switch (unary_op.operation) {
				.boolean_negation => .not,
				.numeric_negation => .neg,
			});
		},

		.invoke => |invoke| {
			const argument_count = try self.compile_argument(environment, invoke.argument);

			try self.compile_expression(environment, invoke.object);
			try self.chunk.opcodes.push_one(.{.call = argument_count});
		},

		.group => |group| try self.compile_expression(environment, group),
		.import_builtin => try self.chunk.opcodes.push_one(.{.push_builtin = .import}),
		.print_builtin => try self.chunk.opcodes.push_one(.{.push_builtin = .print}),
		.vec2_builtin => try self.chunk.opcodes.push_one(.{.push_builtin = .vec2}),
		.vec3_builtin => try self.chunk.opcodes.push_one(.{.push_builtin = .vec3}),

		.declaration_get => |declaration_get| {
			if (get_local_index(environment, declaration_get.declaration)) |index| {
				return self.chunk.opcodes.push_one(.{.push_local = index});
			}

			if (get_capture_index(environment, declaration_get.declaration)) |index| {
				try self.chunk.opcodes.push_one(.{.push_binding = index});

				if (declaration_get.declaration.is.captured) {
					try self.chunk.opcodes.push_one(.get_box);
				}

				return;
			}

			return self.env.raise(error.IllegalState, "local out of scope");
		},

		.declaration_set => |declaration_set| {
			if (get_local_index(environment, declaration_set.declaration)) |index| {
				try self.compile_expression(environment, declaration_set.assign);

				return self.chunk.opcodes.push_one(.{.set_local = index});
			}

			if (get_capture_index(environment, declaration_set.declaration)) |index| {
				try self.chunk.opcodes.push_one(.{.push_binding = index});
				try self.compile_expression(environment, declaration_set.assign);

				if (declaration_set.declaration.is.captured) {
					try self.chunk.opcodes.push_one(.set_box);
				}

				return;
			}

			return self.env.raise(error.IllegalState, "local out of scope");
		},

		.field_get => |field_get| {
			try self.compile_expression(environment, field_get.object);
			try self.chunk.opcodes.push_one(.{.push_const = try self.declare_symbol(field_get.identifier)});
			try self.chunk.opcodes.push_one(.get_dynamic);
		},

		.field_set => |field_set| {
			try self.compile_expression(environment, field_set.object);
			try self.chunk.opcodes.push_one(.{.push_const = try self.declare_symbol(field_set.identifier)});
			try self.compile_expression(environment, field_set.assign);
			try self.chunk.opcodes.push_one(.set_dynamic);
		},

		.subscript_get => |subscript_get| {
			try self.compile_expression(environment, subscript_get.object);
			try self.compile_expression(environment, subscript_get.index);
			try self.chunk.opcodes.push_one(.get_dynamic);
		},

		.subscript_set => |subscript_set| {
			try self.compile_expression(environment, subscript_set.object);
			try self.compile_expression(environment, subscript_set.index);
			try self.compile_expression(environment, subscript_set.assign);
			try self.chunk.opcodes.push_one(.set_dynamic);
		},
	}
}

pub fn compile_environment(self: Self, environment: *const tree.Environment) kym.RuntimeError!void {
	if (environment.statement) |statement| {
		const last_statement = try self.compile_statement(environment, statement);

		if (last_statement.kind != .@"return") {
			try self.chunk.opcodes.push_one(.push_nil);
		}
	}
}

fn compile_statement(self: Self, environment: *const tree.Environment, initial_statement: *const Stmt) kym.RuntimeError!*const Stmt {
	var current_statement = initial_statement;

	while (true) {
		switch (current_statement.kind) {
			.@"return" => |@"return"| {
				if (@"return".returned_expression) |expression| {
					try self.compile_expression(environment, expression);
				} else {
					try self.chunk.opcodes.push_one(.push_nil);
				}
			},

			.@"while" => |@"while"| {
				try self.compile_expression(environment, @"while".loop_expression);
				try self.chunk.opcodes.push_one(.{.jf = 0});

				const origin_index = @as(u32, @intCast(self.chunk.opcodes.values.len - 1));

				_ = try self.compile_statement(environment, @"while".loop);
				self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

				try self.compile_expression(environment, @"while".loop_expression);
				try self.chunk.opcodes.push_one(.{.jt = origin_index});
			},

			.@"if" => |@"if"| {
				try self.compile_expression(environment, @"if".then_expression);
				try self.chunk.opcodes.push_one(.{.jf = 0});

				const origin_index = @as(u32, @intCast(self.chunk.opcodes.values.len - 1));

				_ = try self.compile_statement(environment, @"if".@"then");
				self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

				if (@"if".@"else") |@"else"| {
					_ = try self.compile_statement(environment, @"else");
				}
			},

			.declare => |declare| {
				try self.compile_expression(environment, declare.initial_expression);

				if (declare.declaration.is.captured and !declare.declaration.is.readonly) {
					try self.chunk.opcodes.push_one(.push_boxed);
				}
			},

			.top_expression => |top_expression| {
				try self.compile_expression(environment, top_expression);

				if (top_expression.kind == .invoke) {
					try self.chunk.opcodes.push_one(.pop);
				}
			},
		}

		current_statement = current_statement.next orelse return current_statement;
	}
}

fn declare_chunk(self: Self, chunk: Chunk) kym.RuntimeError!u16 {
	if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
		return self.env.raise(error.BadSyntax, "chunks cannot contain more than 65,535 constants");
	}

	const constant = try self.env.new_dynamic(coral.io.bytes_of(&chunk), Chunk.typeinfo);

	errdefer self.env.discard(constant);

	try self.chunk.constants.push_one(constant);

	return @intCast(self.chunk.constants.values.len - 1);
}

fn declare_fixed(self: Self, fixed: kym.Fixed) kym.RuntimeError!u16 {
	if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
		return self.env.raise(error.BadSyntax, "chunks cannot contain more than 65,535 constants");
	}

	const constant = try self.env.new_fixed(fixed);

	errdefer self.env.discard(constant);

	try self.chunk.constants.push_one(constant);

	return @intCast(self.chunk.constants.values.len - 1);
}

fn declare_float(self: Self, float: kym.Float) kym.RuntimeError!u16 {
	if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
		return self.env.raise(error.BadSyntax, "chunks cannot contain more than 65,535 constants");
	}

	const constant = try self.env.new_float(float);

	errdefer self.env.discard(constant);

	try self.chunk.constants.push_one(constant);

	return @intCast(self.chunk.constants.values.len - 1);
}

fn declare_string(self: Self, string: []const coral.io.Byte) kym.RuntimeError!u16 {
	if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
		return self.env.raise(error.BadSyntax, "chunks cannot contain more than 65,535 constants");
	}

	const constant = try self.env.new_string(string);

	errdefer self.env.discard(constant);

	try self.chunk.constants.push_one(constant);

	return @intCast(self.chunk.constants.values.len - 1);
}

fn declare_symbol(self: Self, symbol: []const coral.io.Byte) kym.RuntimeError!u16 {
	if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
		return self.env.raise(error.BadSyntax, "chunks cannot contain more than 65,535 constants");
	}

	const constant = try self.env.new_symbol(symbol);

	errdefer self.env.discard(constant);

	try self.chunk.constants.push_one(constant);

	return @intCast(self.chunk.constants.values.len - 1);
}

pub fn get_capture_index(self: *const tree.Environment, declaration: *const tree.Declaration) ?u8 {
	if (self.enclosing) |enclosing_environment| {
		var capture_index = @as(u8, 0);

		while (capture_index < self.capture_count) : (capture_index += 1) {
			const captured_declaration_index = self.captures[capture_index];

			coral.debug.assert(captured_declaration_index < enclosing_environment.local_declaration_count);

			if (&enclosing_environment.local_declarations[captured_declaration_index] == declaration) {
				return capture_index;
			}
		}
	}

	return null;
}

pub fn get_local_index(self: *const tree.Environment, declaration: *const tree.Declaration) ?u8 {
	var remaining = self.local_declaration_count;

	while (remaining != 0) {
		remaining -= 1;

		if (&self.local_declarations[remaining] == declaration) {
			return remaining;
		}
	}

	return null;
}
